"""Class representing the internal Packet Filter statistics and counters.

PFStatus objects contain a series of runtime statistical information describing
the current status of the Packet Filter.
"""

import time
from socket import ntohl

from PF._PFStruct import pf_status
from PF.PFConstants import *
from PF.PFUtils import PFObject


__all__ = ['PFStatus']


# Dictionaries for mapping constants to strings ################################
dbg_levels = {LOG_EMERG:   "emerg",
              LOG_ALERT:   "alert",
              LOG_CRIT:    "crit",
              LOG_ERR:     "err",
              LOG_WARNING: "warning",
              LOG_NOTICE:  "notice",
              LOG_INFO:    "info",
              LOG_DEBUG:   "debug"}


# PFStatus class ###############################################################
class PFStatus(PFObject):
    """Class representing the internal Packet Filter statistics and counters."""

    _struct_type = pf_status

    def __init__(self, status):
        """Check argument and initialize class attributes."""
        super(PFStatus, self).__init__(status)

    def _from_struct(self, s):
        """Initialize class attributes from a pf_status structure."""
        self.ifname    = s.ifname
        self.running   = bool(s.running)
        #stateid
        self.since     = s.since
        self.states    = s.states
        self.src_nodes = s.src_nodes
        self.debug     = s.debug
        self.hostid    = ntohl(s.hostid) & 0xffffffff
        self.reass     = s.reass
        self.pf_chksum = "0x" + "".join(map("{0:02x}".format, s.pf_chksum))

        self.cnt       = {'match':                    s.counters[0],
                          'bad-offset':               s.counters[1],
                          'fragment':                 s.counters[2],
                          'short':                    s.counters[3],
                          'normalize':                s.counters[4],
                          'memory':                   s.counters[5],
                          'bad-timestamp':            s.counters[6],
                          'congestion':               s.counters[7],
                          'ip-option':                s.counters[8],
                          'proto-cksum':              s.counters[9],
                          'state-mismatch':           s.counters[10],
                          'state-insert':             s.counters[11],
                          'state-limit':              s.counters[12],
                          'src-limit':                s.counters[13],
                          'synproxy':                 s.counters[14]}

        self.lcnt      = {'max states per rule':      s.lcounters[0],
                          'max-src-states':           s.lcounters[1],
                          'max-src-nodes':            s.lcounters[2],
                          'max-src-conn':             s.lcounters[3],
                          'max-src-conn-rate':        s.lcounters[4],
                          'overload table insertion': s.lcounters[5],
                          'overload flush states':    s.lcounters[6]}

        self.fcnt      = {'searches':                 s.fcounters[0],
                          'inserts':                  s.fcounters[1],
                          'removals':                 s.fcounters[2]}

        self.scnt      = {'searches':                 s.scounters[0],
                          'inserts':                  s.scounters[1],
                          'removals':                 s.scounters[2]}

        self.bytes     = {'in':   (s.bcounters[0][0], s.bcounters[1][0]),
                          'out':  (s.bcounters[0][1], s.bcounters[1][1])}

        self.packets   = {'in':  ((s.pcounters[0][0][PF_PASS],
                                   s.pcounters[1][0][PF_PASS]),
                                  (s.pcounters[0][0][PF_DROP],
                                   s.pcounters[1][0][PF_DROP])),
                          'out': ((s.pcounters[0][1][PF_PASS],
                                   s.pcounters[1][1][PF_PASS]),
                                  (s.pcounters[0][1][PF_DROP],
                                   s.pcounters[1][1][PF_DROP]))}

    def _to_string(self):
        """Return a string containing the statistics."""
        s = "Status: " + ('Enabled' if self.running else 'Disabled')

        if self.since:
            runtime = int(time.time()) - self.since
            day, sec = divmod(runtime, 60)
            day, min = divmod(day, 60)
            day, hrs = divmod(day, 24)
            s += " for {0} days {1:02}:{2:02}:{3:02}".format(day, hrs, min, sec)

        dbg = dbg_levels.get(self.debug, "unknown")
        s = "{0:<44}{1:>15}\n\n".format(s, "Debug: " + dbg)
        s += "Hostid:   0x{0.hostid:08x}\n".format(self)
        s += "Checksum: {0.pf_chksum}\n\n".format(self)

        if self.ifname:
            fmt = "  {0:<25} {1[0]:>14d} {1[1]:>16d}\n"
            s += "Interface Stats for {0.ifname:<16} ".format(self)
            s += "{0:>5} {1:>16}\n".format("IPv4", "IPv6")
            s += fmt.format("Bytes In", self.bytes["in"])
            s += fmt.format("Bytes Out", self.bytes["out"])
            s += "  Packets In\n"
            s += fmt.format("  Passed", self.packets["in"][PF_PASS])
            s += fmt.format("  Blocked", self.packets["in"][PF_DROP])
            s += "  Packets Out\n"
            s += fmt.format("  Passed", self.packets["out"][PF_PASS])
            s += fmt.format("  Blocked", self.packets["out"][PF_DROP])
            s += "\n"

        s += "{0:<27} {1:>14} {2:>16}\n".format("State Table", "Total", "Rate")
        s += "  {0:<25} {1.states:>14d}".format("current entries", self)
        for k, v in self.fcnt.iteritems():
            s += "\n  {0:<25} {1:>14d} ".format(k, v)
            if self.since:
                s += "{0:>14.1f}/s".format(float(v)/runtime)

        s += "\nSource Tracking Table\n"
        s += "  {0:<25} {1.src_nodes:>14d}".format("current entries", self)
        for k, v in self.scnt.iteritems():
            s += "\n  {0:<25} {1:>14d} ".format(k, v)
            if self.since:
                s += "{0:>14.1f}/s".format(float(v)/runtime)

        s += "\nCounters"
        for k, v in self.cnt.iteritems():
            s += "\n  {0:<25} {1:>14d} ".format(k, v)
            if self.since:
                s += "{0:>14.1f}/s".format(float(v)/runtime)

        s += "\nLimit Counters"
        for k, v in self.lcnt.iteritems():
            s += "\n  {0:<25} {1:>14d} ".format(k, v)
            if self.since:
                s += "{0:>14.1f}/s".format(float(v)/runtime)

        return s
